REBOL [
	Name: "Async CALL"
	Date: 28/08/2004
	Version: 1.1
	Author: "Nenad Rakocevic / SOFTINNOV"
	Purpose: "Asynchronous execution of external commands"
	History: [
		1.1  - 28/08/2004 - {
			o Fixed a bug that keeps the port busy when the external process has 
			  finished. Thanks to Graham Chiu for reporting.
			o Documentation examples fixed.
		}
		1.0  - 17/02/2004 - {
			o First public release.
			o Documentation and commented version added.
		}
		0.91 - 11/02/2004 -	{
			o Bugfix: a buffer overflow was causing some output bytes to be missed.
			  Thanks to Gabriele Santilli for reporting the bug.
			o Code minor clean-up.
		}
		0.9  - 10/02/2004 -	"Beta release"
	]
	License: {
		(C) 2004 SOFTINNOV

		This software is provided 'as-is', without any express or implied
		warranty.  In no event will the author be held liable for any damages
		arising from the use of this software.

		Permission is granted to anyone to use this software for any purpose,
		including commercial applications, and to alter it and redistribute it
		freely, subject to the following restrictions:

		1. The origin of this software must not be misrepresented; you must not
		claim that you wrote the original software. If you use this software
		in a product, an acknowledgment in the product documentation would be
		appreciated but is not required.

		2. Altered source versions must be plainly marked as such, and must not be
		misrepresented as being the original software.

		3. This notice may not be removed or altered from any source distribution.

		Nenad Rakocevic
		nr@softinnov.com
	}
]

OS: context [

	win-kernel: load/library %kernel32.dll

	SECURITY_ATTRIBUTES: make struct! [
		nLength 			 [integer!]
		lpSecurityDescriptor [integer!]
		bInheritHandle 		 [integer!]
	] none

	STARTUPINFO: make struct! startup-info-struct: [
		cb 				[integer!]
		lpReserved 		[integer!]
		lpDesktop		[integer!]
		lpTitle			[integer!]
		dwX				[integer!]
		dwY				[integer!]
		dwXSize			[integer!]
		dwYSize			[integer!]
		dwXCountChars 	[integer!]
		dwYCountChars 	[integer!]
		dwFillAttribute	[integer!]
		dwFlags			[integer!]
		wShowWindow		[short]
		cbReserved2		[short]
		lpReserved2		[integer!]
		hStdInput		[integer!]
		hStdOutput		[integer!]
		hStdError		[integer!]
	] none

	PROCESS_INFORMATION: make struct! [
		hProcess	[integer!]
		hThread 	[integer!]
		dwProcessID	[integer!]
		dwThreadID	[integer!]
	] none

	CreatePipe: make routine!  [
		phReadPipe 		 [struct! [num [integer!]]]
		phWritePipe 	 [struct! [num [integer!]]]
		lpPipeAttributes [struct! [a [integer!] b [integer!] c [integer!]]]
		nSize 			 [integer!]
		return:			 [integer!]
	] win-kernel "CreatePipe"

	ReadFile: make routine! [
		hFile 				 [integer!]
		lpBuffer 			 [string!]
		nNumberOfBytesToRead [integer!]
		lpNumberOfBytesRead  [struct! [num [integer!]]]
		lpOverlapped 		 [integer!]
		return:				 [integer!]
	] win-kernel "ReadFile"

	CreateProcess: make routine! compose/deep [
		lpApplicationName	 [integer!]
		lpCommandLine		 [string!]	
		lpProcessAttributes	 [struct! [a [integer!] b [integer!] c [integer!]]]
		lpThreadAttributes	 [struct! [a [integer!] b [integer!] c [integer!]]]
		bInheritHandles		 [char!]
		dwCreationFlags		 [integer!]
		lpEnvironment		 [integer!]
		lpCurrentDirectory	 [integer!]
		lpStartupInfo		 [struct! [(startup-info-struct)]]
		lpProcessInformation [struct! [a [integer!] b [integer!] c [integer!] d [integer!]]]
		return:				 [integer!]
	] win-kernel "CreateProcessA"

	CloseHandle: make routine! [
		hObject	[integer!]
		return: [integer!]
	] win-kernel "CloseHandle"

	PeekNamedPipe: make routine! [
		hNamedPipe			[integer!]
		lpBuffer			[integer!]
		nBufferSize			[integer!]
		lpBytesRead			[integer!]
		lpTotalBytesAvail	[struct! [cnt [integer!]]]
		lpBytesLeftThisMessage [integer!]
		return: 			[integer!]
	] win-kernel "PeekNamedPipe"

	GetExitCodeProcess: make routine! [
		hProcess	[integer!]
		lpExitCode	[struct! [int [integer!]]] 
		return:		[integer!]
	] win-kernel "GetExitCodeProcess"

	NORMAL_PRIORITY_CLASS:	to-integer #{00000020}
	STARTF_USESTDHANDLES: 	to-integer #{00000100}
	STARTF_USESHOWWINDOW: 	1
	SW_HIDE: 				0
	STILL_ACTIVE:			259

	null: to-char 0
	lpDWORD: make struct! [int [integer!]] none
	make-lpDWORD: does [make struct! lpDWORD [0]]
	
	sa: make struct! SECURITY_ATTRIBUTES [0 0 1]
	sa/nLength: length? third sa
	
	start-info: make struct! STARTUPINFO none
	start-info/cb: length? third start-info
	start-info/dwFlags: STARTF_USESTDHANDLES 
]

make root-protocol [
	scheme: 'call
	
	copy*: 	 get in system/words 'copy
	insert*: get in system/words 'insert
	pick*: 	 get in system/words 'pick
	close*:  get in system/words 'close
	net-log: get in net-utils	 'net-log

	open-proto: none
	port-flags: system/standard/port-flags/pass-thru

	events-autofire: func [port state [logic!]][
		port/sub-port/state/inbuffer: pick reduce ["o" none] state
	]
	
	init-locals: func [port][
		port/locals: context [
			on-input: 
			on-completed: none
			show-window: no
			output: make string! 1024
			pipe-buffer: make string! 256
			insert*/dup pipe-buffer OS/null 256	

			start-info: make struct! OS/start-info second OS/start-info
			proc-info: make struct! OS/PROCESS_INFORMATION none

			read-handle:  OS/make-lpDWORD
			write-handle: OS/make-lpDWORD
			bytes-avail:  OS/make-lpDWORD
			bytes-read:   OS/make-lpDWORD
			exit-code:    OS/make-lpDWORD
		]
	]

	launch-call: func [port cmd-line [string!] /local pl ret][
		pl: port/locals
		clear pl/output
		cmd-line: join cmd-line OS/null
		change/dup pl/pipe-buffer OS/null 256
		pl/proc-info/hProcess: -1	; to avoid premature exit in awake
		
		ret: catch [
			ret: OS/CreatePipe pl/read-handle pl/write-handle OS/sa 0			
			if zero? ret [throw 1]

			if not pl/show-window [
				pl/start-info/dwFlags: pl/start-info/dwFlags or OS/STARTF_USESHOWWINDOW
			]
			pl/start-info/hStdOutput: pl/write-handle/int
			pl/start-info/hStdError: pl/write-handle/int
		
			ret: OS/CreateProcess 0 cmd-line OS/sa OS/sa to-char 1 0 0 0 
				pl/start-info pl/proc-info
			if zero? ret [throw 2]
			
			ret: none
		]
		if integer? ret [
			make error! join pick ["CreatePipe" "CreateProcess"] ret " failed!"
		]
		events-autofire port on
		none
	]

;------- External API -------
	
    init: func [port spec][
        port/url: spec
        port/target: any [port/target "call"]
    ]
    open: func [port][    
        port/sub-port: make port! [
        	scheme: 'call
        	host: target: ""
        ]      
        ;-- Note: [make port!] set state to none
        port/sub-port/state: context [
			flags: 524803 or system/standard/port-flags/pass-thru
			misc: [320 [] 0]
			tail: num: 0
			with: custom: none
			index: 0
			func: 3
			fpos: 0
			inBuffer: outBuffer: none
		]
		port/state/flags: port/state/flags or port-flags
		init-locals port
    ]
    insert: func  [port data [string!]][
    	if empty? data [
    		make error! "no command to execute!"
    	]
    	either query port [
    		make error! "port already busy!"
    	][
    		launch-call port data
    	]
    ]
    copy: func [port][
    	make error! "not supported"
    ]
    close: func [port][
        port/state: port/sub-port/state
        port/state/flags: 0
    ]
    write: func [port data][
        make error! "not supported"
    ]
    read: func [port data][
       make error! "not supported"
    ]
    get-sub-port: func [port /local pl ret remain][
        pl: port/locals		
		if not zero? pl/proc-info/hProcess [
			ret: catch [
				ret: OS/GetExitCodeProcess pl/proc-info/hProcess pl/exit-code
				if zero? ret [throw 3]
				
				ret: OS/PeekNamedPipe pl/read-handle/int 0 0 0 pl/bytes-avail 0
				if zero? ret [throw 1]

				if not zero? remain: pl/bytes-avail/int [ 
					until [
						ret: OS/ReadFile pl/read-handle/int pl/pipe-buffer
							length? pl/pipe-buffer pl/bytes-read 0	

						if zero? ret [throw 2]

						append pl/output ret: copy*/part pl/pipe-buffer pl/bytes-read/int
							pl/on-input port ret

						change/dup pl/pipe-buffer OS/null 256
						remain: remain - pl/bytes-read/int 
						zero? remain 
					]
				]									
				if pl/exit-code/int <> OS/STILL_ACTIVE [
					;events-autofire port off
					OS/CloseHandle pl/proc-info/hProcess
					OS/CloseHandle pl/proc-info/hThread
					OS/CloseHandle pl/read-handle/int
					OS/CloseHandle pl/write-handle/int
					pl/proc-info/hProcess: 0
					pl/on-completed port
				]
				ret: none
			]
			if integer? ret [
				make error! join pick [
					"PeekNamedPipe"
					"ReadFile"
					"GetExitCodeProcess"
				] ret " failed!"
			]
		]		
        port/sub-port
    ]
    awake: func [port][
		either zero? port/locals/proc-info/hProcess [	
			events-autofire port off
			not find system/ports/wait-list port
		][
			false
		]
    ]
    query: func [port][to-logic port/sub-port/state/inbuffer]
    
   	net-utils/net-install CALL self none
]