REBOL [
	Name: "Async CALL Ping Demo"
	Date: 11/02/2003
	Version: 1.0
	Author: "Nenad Rakocevic / SOFTINNOV"
	Purpose: "Demonstrate the usage of call://"
	Notes: {
		This demo script will show how to mix call:// ports with View
		by executing 3 ping commands simultaneously.
	}
]

;-- Install the call:// scheme
do %async-call.r

view/title layout [
	style log-area area 450x150 font [name: "Courrier New" size: 9]
	style host field 150
	button "Ping all" [
		;-- Test if the first call port is busy
		if query clist/1 [
			alert "You're not allowed to do that, ports are already busy !"
			return
		]		
		repeat i 3 [
			clear flist/:i/line-list
			clear flist/:i/text
			flist/:i/colors/1: white		
			;-- Send to each ports a ping command on different hosts.
			insert clist/:i join "ping " hlist/:i/text
		]
		;-- refresh all log areas
		show flist
		;-- reset the job counter
		end-cnt: 3
		;-- starts moving the bouncing box
		bounce/rate: 20
		show bounce
	]
	bounce: box pink "I can move while the calls are executed" edge [
		size: 1x1
	] font [
		color: aqua
		style: shadow: none
	] feel [	
		engage: func [f a e][
			if a = 'time [
				f/offset/y: f/offset/y + (8 * dir)
				if any [f/offset/y < 53 f/offset/y > 203][dir: negate dir]
				show f
			]
		]
	]
	return
	
	host1: host "www.rebol.com" 	log1: log-area
	host2: host "www.nasa.gov"		log2: log-area
	host3: host "www.wanadoo.fr"	log3: log-area
	
	do [
		;--Initialization code
		
		;-- Moving box vertical direction (1 or -1)
		dir: 1 
		;-- Finish jobs counter
		end-cnt: 3
		;-- Call:// ports list
		clist: make block! 3
		;-- Logging area faces list
		flist: reduce [log1 log2 log3]
		;-- Host URL faces list
		hlist: reduce [host1 host2 host3]
		
		repeat i 3 [
			new: open call://
			
			new/locals: make new/locals [
				;-- Add a reference to its logging face
				area: pick flist i
				on-input: func [port data][			
					;-- Append the new data to the logging face and refresh it
					clear port/locals/area/line-list
					append port/locals/area/text data
					show port/locals/area
				]
				on-completed: func [port][
					;-- If all jobs are finished
					if zero? end-cnt: end-cnt - 1 [
						;-- Stop moving the box
						bounce/rate: none
						show bounce
					]
					;-- Signal that the job is finished by changing the bg color
					port/locals/area/colors/1: green
					show port/locals/area
				]
			]
			append clist new
		]
		;-- Add the opened call ports to the system port list
		;-- This allows REBOL to process View *and* call port events.
		append system/ports/wait-list clist
	]
] "Multiple ping demo"

quit